from contextlib import contextmanager
from typing import Callable, ContextManager

from loguru import logger
from sqlalchemy.exc import OperationalError
from sqlalchemy.orm import Session


def sql_session_scope_provider(
    session_factory: Callable[[], Session]
) -> Callable[[], ContextManager[Session]]:
    @contextmanager
    def session_scope() -> ContextManager[Session]:
        session = session_factory()
        try:
            yield session
            session.commit()
        except OperationalError as e:
            logger.error(e)
            session.rollback()
            raise e
        except Exception as e:
            logger.error(e)
            session.rollback()
            raise e
        finally:
            session.close()
        # session.delete()

    return session_scope



from __future__ import annotations

import inspect
from abc import abstractmethod
from typing import Generic, TypeVar

from sqlalchemy.orm import DeclarativeBase

T = TypeVar("T")


class SqlBase(DeclarativeBase, Generic[T]):
    def __repr__(self):
        attributes = ", ".join(
            f"{key}={value}"
            for key, value in self.__dict__.items()
            if not key.startswith("_")
        )
        return f"{self.__class__.__name__}({attributes})"

    def info(self) -> dict[str, str]:
        return {
            "name": self.__class__.__name__,
            "module": self.__class__.__module__,
            "file": inspect.getsourcefile(self.__class__),
        }

    @abstractmethod
    def to_domain(self) -> T:
        pass

    @staticmethod
    @abstractmethod
    def from_domain(domain_entity: T) -> SqlBase:
        pass



from typing import Any

from petisco.base.domain.message.message import Message


class DomainEvent(Message):
    def model_post_init(self, __context: Any) -> None:
        self._message_type = "domain_event"
        super().model_post_init(__context)



from petisco.base.domain.persistence.databases import databases
from petisco.extra.sqlalchemy.sql.mysql.mysql_connection import MySqlConnection


def mysql_is_running_locally() -> bool:
    try:
        from petisco.extra.sqlalchemy.sql.sql_database import SqlDatabase

        database = SqlDatabase(connection=MySqlConnection.create_local())
        databases.add(database)
        databases.initialize()
        is_running_locally = True
    except:  # noqa: E722
        is_running_locally = False
    databases.remove(SqlDatabase, skip_if_not_exist=True)
    return is_running_locally



from petisco.extra.logger.log_message import LogMessage
from petisco.extra.logger.logger import Logger


class NotImplementedLogger(Logger):
    def log(self, logging_level, log_message: LogMessage):
        pass



from fastapi import FastAPI
from fastapi.routing import APIRoute

from petisco.extra.fastapi.application.response_mocker import ResponseMocker


def add_controller_responses_to_response_mocker_dependencies(app: FastAPI) -> None:
    for route in app.routes:
        if isinstance(route, APIRoute) and route.responses:
            for depends in route.dependencies:
                if isinstance(depends.dependency, ResponseMocker):
                    depends.dependency.responses = route.responses



from typing import List, Union

from petisco.base.application.dependency_injection.dependency import Dependency
from petisco.base.domain.message.command_bus import CommandBus
from petisco.base.domain.message.domain_event_bus import DomainEventBus
from petisco.base.domain.message.message_configurer import MessageConfigurer
from petisco.base.domain.message.message_consumer import MessageConsumer
from petisco.base.domain.message.not_implemented_command_bus import (
    NotImplementedCommandBus,
)
from petisco.base.domain.message.not_implemented_domain_event_bus import (
    NotImplementedDomainEventBus,
)
from petisco.base.domain.message.not_implemented_message_comsumer import (
    NotImplementedMessageConsumer,
)
from petisco.base.domain.message.not_implemented_message_configurer import (
    NotImplementedMessageConfigurer,
)
from petisco.base.misc.builder import Builder


def get_default_message_dependencies() -> List[Dependency]:
    return [
        Dependency(
            DomainEventBus,
            builders={"default": Builder(NotImplementedDomainEventBus)},
            envar_modifier="PETISCO_DOMAIN_EVENT_BUS_TYPE",
        ),
        Dependency(
            CommandBus,
            builders={"default": Builder(NotImplementedCommandBus)},
            envar_modifier="PETISCO_COMMAND_BUS_TYPE",
        ),
        Dependency(
            MessageConfigurer,
            builders={"default": Builder(NotImplementedMessageConfigurer)},
            envar_modifier="PETISCO_MESSAGE_CONFIGURER_TYPE",
        ),
        Dependency(
            MessageConsumer,
            builders={"default": Builder(NotImplementedMessageConsumer)},
            envar_modifier="PETISCO_MESSAGE_CONSUMER_TYPE",
        ),
    ]


def get_rabbitmq_message_dependencies(
    organization: str,
    service: str,
    max_retries: int = 5,
    alias: Union[str, None] = None,
) -> List[Dependency]:
    from petisco.extra.rabbitmq.application.message.bus.rabbitmq_command_bus import (
        RabbitMqCommandBus,
    )
    from petisco.extra.rabbitmq.application.message.bus.rabbitmq_domain_event_bus import (
        RabbitMqDomainEventBus,
    )
    from petisco.extra.rabbitmq.application.message.configurer.rabbitmq_message_configurer import (
        RabbitMqMessageConfigurer,
    )
    from petisco.extra.rabbitmq.application.message.consumer.rabbitmq_message_consumer import (
        RabbitMqMessageConsumer,
    )

    return [
        Dependency(
            DomainEventBus,
            alias=alias,
            builders={
                "default": Builder(
                    RabbitMqDomainEventBus, organization=organization, service=service
                ),
                "not_implemented": Builder(NotImplementedDomainEventBus),
            },
            envar_modifier="PETISCO_DOMAIN_EVENT_BUS_TYPE",
        ),
        Dependency(
            CommandBus,
            alias=alias,
            builders={
                "default": Builder(
                    RabbitMqCommandBus, organization=organization, service=service
                ),
                "not_implemented": Builder(NotImplementedCommandBus),
            },
            envar_modifier="PETISCO_COMMAND_BUS_TYPE",
        ),
        Dependency(
            MessageConfigurer,
            alias=alias,
            builders={
                "default": Builder(
                    RabbitMqMessageConfigurer,
                    organization=organization,
                    service=service,
                ),
                "not_implemented": Builder(NotImplementedMessageConfigurer),
            },
            envar_modifier="PETISCO_MESSAGE_CONFIGURER_TYPE",
        ),
        Dependency(
            MessageConsumer,
            alias=alias,
            builders={
                "default": Builder(
                    RabbitMqMessageConsumer,
                    organization=organization,
                    service=service,
                    max_retries=max_retries,
                ),
                "not_implemented": Builder(NotImplementedMessageConsumer),
            },
            envar_modifier="PETISCO_MESSAGE_CONSUMER_TYPE",
        ),
    ]



from functools import wraps
from inspect import signature
from typing import Any, Callable

import elasticapm
from loguru import logger
from meiga import Error, Failure
from meiga.on_failure_exception import OnFailureException

from petisco.base.domain.errors.unknown_error import UnknownError
from petisco.base.misc.result_mapper import ResultMapper
from petisco.base.misc.wrapper import get_middleware_instances


def async_wrapper(
    execute_func: Callable[..., Any],
    wrapped_class_name: str,
    config: Any,
    mapper: ResultMapper,
) -> Callable[..., Any]:
    @wraps(execute_func)
    async def wrapped(*args: Any, **kwargs: Any) -> Any:
        middlewares = get_middleware_instances(config)
        arguments = signature(execute_func).bind(*args, **kwargs).arguments
        arguments.pop("self")

        for middleware in middlewares:
            try:
                middleware.set_data(wrapped_class_name, arguments)
                middleware.before()
            except Exception as exception:
                logger.error(f"Error in the {wrapped_class_name} middlewares (before).")
                logger.exception(exception)

        try:
            result = await execute_func(*args, **kwargs)
        except OnFailureException as exc:
            result = exc.result
        except Error as error:
            result = Failure(error)
        except Exception as exception:
            unknown_error = UnknownError.from_exception(
                exception=exception,
                arguments=arguments if len(arguments) > 0 else args,
                class_name=wrapped_class_name,
            )
            result = Failure(unknown_error)
            client = elasticapm.get_client()
            if client:
                client.capture_exception()

        for middleware in middlewares:
            if result:
                try:
                    middleware.after(result)
                except Exception as exception:
                    logger.error(
                        f"Error in the {wrapped_class_name} middlewares (after)."
                    )
                    logger.exception(exception)

        try:
            result.set_transformer(mapper.map)
        except AttributeError:  # noqa
            raise TypeError(
                f"Controller Error: Return value `{result}` ({type(result)}) must be a `meiga.Result` to "
                f"transform values to success and failure handlers."
            )
        return result

    return wrapped
